# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import pyworkflow.utils as pwutils
import pwem.protocols as emprot
from pyworkflow.tests import BaseTest, setupTestProject

from . import DataSet
import tomoviz.protocols
import tomo.protocols


# class TestTomoPickingConsenus(BaseTest):
#     """This class check if the consensus from a series of SetOfCoordinates3D works properly."""
#
#     @classmethod
#     def setUpClass(cls):
#         setupTestProject(cls)
#         cls.dataset = DataSet.getDataSet('reliontomo')
#         cls.tomogram_1 = cls.dataset.getFile('tomo1')
#         cls.tomogram_2 = cls.dataset.getFile('tomo2')
#
#     def _importSetOfCoordinates(self, tomoFile, pattern):
#         protImportTomogram = self.newProtocol(tomo.protocols.ProtImportTomograms,
#                                               filesPath=tomoFile,
#                                               samplingRate=1,
#                                               objLabel=pwutils.removeBaseExt(tomoFile))
#
#         self.launchProtocol(protImportTomogram)
#
#         output = getattr(protImportTomogram, 'outputTomograms', None)
#
#         self.assertIsNotNone(output,
#                              "There was a problem with tomogram output")
#
#         protImportCoordinates3d = self.newProtocol(tomo.protocols.ProtImportCoordinates3D,
#                                   objLabel='Coordinates ' + pwutils.removeBaseExt(tomoFile),
#                                   auto=tomo.protocols.ProtImportCoordinates3D.IMPORT_FROM_AUTO,
#                                   filesPath=self.dataset.getPath(),
#                                   importTomograms=protImportTomogram.outputTomograms,
#                                   filesPattern=pattern, boxSize=32,
#                                   samplingRate=1)
#         self.launchProtocol(protImportCoordinates3d)
#
#         return getattr(protImportCoordinates3d, 'outputCoordinates', None)
#
#     def _runTomoPickingConsensus(self, coordinates, consensus, mode, radius=100):
#         choices_mode = ['>=', '=']
#         choices_consensus = {-1: 'AND', 1: 'OR'}
#         protPickingConsensus = self.newProtocol(tomoviz.protocols.ProtTomoConsensusPicking,
#                                                 inputCoordinates=coordinates,
#                                                 consensus=consensus,
#                                                 mode=mode,
#                                                 consensusRadius=radius,
#                                                 objLabel='Consenus - ' + choices_consensus[consensus]
#                                                           + ' - ' + choices_mode[mode])
#         self.launchProtocol(protPickingConsensus)
#         return protPickingConsensus
#
#     def test_picking_consenus(self):
#         coords = []
#         coords.append(self._importSetOfCoordinates(self.tomogram_1, '*.tbl'))
#         coords.append(self._importSetOfCoordinates(self.tomogram_2, '*.tbl'))
#
#         # AND + >=
#         protConsensus = self._runTomoPickingConsensus(coords, -1, 0)
#         output = getattr(protConsensus, 'consensusCoordinates', None)
#         self.assertTrue(output,
#                              "There was a problem with consenus output")
#         self.assertTrue(output.getSize() == 20)
#         self.assertTrue(output.getBoxSize() == 32)
#         self.assertTrue(output.getSamplingRate() == 1)
#
#         # AND + =
#         protConsensus = self._runTomoPickingConsensus(coords, -1, 1, radius=50)
#         output = getattr(protConsensus, 'consensusCoordinates', None)
#         self.assertTrue(output,
#                              "There was a problem with consenus output")
#         self.assertTrue(output.getSize() == 5)
#         self.assertTrue(output.getBoxSize() == 32)
#         self.assertTrue(output.getSamplingRate() == 1)
#
#         # OR + >=
#         protConsensus = self._runTomoPickingConsensus(coords, 1, 0)
#         output = getattr(protConsensus, 'consensusCoordinates', None)
#         self.assertTrue(output,
#                              "There was a problem with consenus output")
#         self.assertTrue(output.getSize() == 83)
#         self.assertTrue(output.getBoxSize() == 32)
#         self.assertTrue(output.getSamplingRate() == 1)
#
#         # OR + =
#         protConsensus = self._runTomoPickingConsensus(coords, 1, 1)
#         output = getattr(protConsensus, 'consensusCoordinates', None)
#         self.assertTrue(output,
#                              "There was a problem with consenus output")
#         self.assertTrue(output.getSize() == 63)
#         self.assertTrue(output.getBoxSize() == 32)
#         self.assertTrue(output.getSamplingRate() == 1)
#
#         return output
#
# class TestTomoPickingRemoveDuplicates(BaseTest):
#     """This class check if the removal of duplicates from a series
#      of SetOfCoordinates3D works properly."""
#
#     @classmethod
#     def setUpClass(cls):
#         setupTestProject(cls)
#         cls.dataset = DataSet.getDataSet('reliontomo')
#         cls.tomogram_1 = cls.dataset.getFile('tomo1')
#         cls.tomogram_2 = cls.dataset.getFile('tomo2')
#
#     def _importSetOfCoordinates(self, tomoFile, pattern):
#         protImportTomogram = self.newProtocol(tomo.protocols.ProtImportTomograms,
#                                               filesPath=tomoFile,
#                                               samplingRate=5,
#                                               objLabel=pwutils.removeBaseExt(tomoFile))
#
#         self.launchProtocol(protImportTomogram)
#
#         output = getattr(protImportTomogram, 'outputTomograms', None)
#
#         self.assertIsNotNone(output,
#                              "There was a problem with tomogram output")
#
#         protImportCoordinates3d = self.newProtocol(tomo.protocols.ProtImportCoordinates3D,
#                                   objLabel='Coordinates ' + pwutils.removeBaseExt(tomoFile),
#                                   auto=tomo.protocols.ProtImportCoordinates3D.IMPORT_FROM_AUTO,
#                                   filesPath=self.dataset.getPath(),
#                                   importTomograms=protImportTomogram.outputTomograms,
#                                   filesPattern=pattern, boxSize=32,
#                                   samplingRate=5)
#         self.launchProtocol(protImportCoordinates3d)
#
#         return getattr(protImportCoordinates3d, 'outputCoordinates', None)
#
#     def _joinSets(self, sets):
#         protJoinSets = self.newProtocol(emprot.ProtUnionSet,
#                                         inputSets=sets,
#                                         objLabel='Duplicated Coordinates 3D')
#         self.launchProtocol(protJoinSets)
#         return getattr(protJoinSets, 'outputSet', None)
#
#     def _runTomoPickingRemoveDuplicates(self, coordinates):
#         protPickingConsensus = self.newProtocol(tomoviz.protocols.ProtTomoPickingRemoveDuplicates,
#                                                 inputCoordinates=coordinates,
#                                                 objLabel='Remove Duplicates')
#         self.launchProtocol(protPickingConsensus)
#         return protPickingConsensus
#
#     def test_picking_remove_duplicates(self):
#         coords = []
#         coords.append(self._importSetOfCoordinates(self.tomogram_1, '*.tbl'))
#         coords.append(self._importSetOfCoordinates(self.tomogram_1, '*.tbl'))
#         all_coords = self._joinSets(coords)
#         protRemoveDuplicates = self._runTomoPickingRemoveDuplicates(all_coords)
#         output = getattr(protRemoveDuplicates, 'outputCoordinates', None)
#         self.assertTrue(output,
#                              "There was a problem with consenus output")
#         self.assertTrue(output.getSize() == 63)
#         self.assertTrue(output.getBoxSize() == 32)
#         self.assertTrue(output.getSamplingRate() == 5)
#
#         return output


if __name__ == 'main':
    pass



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import pyvista as pv
import numpy as np
from scipy.spatial.distance import pdist
from multiprocessing import Process

from PyQt5.QtWidgets import QApplication, QMainWindow, QDockWidget
from pyvistaqt.plotting import QtInteractor
from PyQt5.QtCore import Qt

class TriangulationPlot(object):
    '''
    Class to visualize triangulation(s) and/or point cloud(s) with their associated normals
    Input paramters:
         - meshes (Mandatory):    List containing the pyvista triangulations to be shown by the viewer
         - clouds (Optional):     List containing the original point clouds associated to the triangulations.
                                  This parameter is useful to compare the coordinates adjusted to the mesh and
                                  the original ones
         -extNormals (Optional):  List containing the original normals associated to each vertex in the
                                  Delaunay triangulation. In general, this parameter is used to compare the
                                  normals computed inside Scipion (Pyvista normals stored in the meshes objects)
                                  and those normals obtained with other software (PySeg, Dynamo...)
    Usage:
         import TriangulationPlot
         plt = TriangulationPlot([mesh], clouds=[cloud], extNormals=[extNormal])
         plt.initializePlot()
    '''


    def __init__(self, meshes, clouds=None, extNormals_List=None, extNormals_coords=None):
        self.clouds = clouds
        self.meshes = meshes
        self.extNormals_List = extNormals_List
        self.extNormals_coords = extNormals_coords
        self.initialize_PyQt()
        self.actor_meshes = []
        self.actor_normals = []
        self.actor_extNormals = []
        self.actor_points = []
        self.actor_ori_points = []
        self.buttonMesh = self.p.add_checkbox_button_widget(callback=self.plotMesh, position=(65., 10.),
                                                            color_on='white')
        self.p.add_text('Mesh', position=(10.+48, 65.), font_size=12)
        self.p.add_text('Normals', position=(200.+44, 65.), font_size=12)
        self.buttonNormals = self.p.add_checkbox_button_widget(callback=self.plotNormals, position=(265., 10.))
        self.p.add_text('Cloud', position=(420.+33, 65.), font_size=12)
        self.buttonPoints = self.p.add_checkbox_button_widget(callback=self.plotPoints, position=(465., 10.),
                                                              color_on='black')
        if self.clouds is not None:
            self.p.add_text('Original Cloud', position=(598., 65.), font_size=12)
            self.buttonOriPoints = self.p.add_checkbox_button_widget(callback=self.plotOriPoints, position=(665., 10.),
                                                                     color_on='white')
        if self.extNormals_List is not None:
            self.p.add_text('External Normals', position=(598.+198, 65.), font_size=12)
            self.buttonExtNormals = self.p.add_checkbox_button_widget(callback=self.plotExtNormals, position=(865., 10.),
                                                                      color_on='red')

        self.mainWindow.show()

    def initialize_PyQt(self):
        self.app = QApplication([])
        self.mainWindow = QMainWindow()
        self.mainWindow.default_style_sheet = self.mainWindow.styleSheet()
        self.mainWindow.save_commands = True  # stores commands internally when True
        self.mainWindow.hold = False
        self.mainWindow.load_dialog = None
        self.mainWindow.app = self.app
        self.mainWindow.resize(1000, 600)
        self.p = QtInteractor(self.mainWindow)
        self.p.app_window = self.mainWindow
        self.mainWindow.dock_vtk = QDockWidget('Mesh Viewer', self.mainWindow)
        self.mainWindow.dock_vtk.setWidget(self.p.interactor)
        self.mainWindow.dock_vtk.setMinimumSize(700, 600)
        self.mainWindow.addDockWidget(Qt.LeftDockWidgetArea, self.mainWindow.dock_vtk)
        self.p.add_toolbars()

    def plotMesh(self, value):
        if value:
            for mesh in self.meshes:
                self.actor_meshes.append(self.p.add_mesh(mesh, show_edges=True))
            self.buttonPoints.GetRepresentation().SetState(False)
            if self.buttonOriPoints.GetRepresentation().GetState:
                for actor in self.actor_ori_points:
                    self.p.remove_actor(actor)
                self.buttonOriPoints.GetRepresentation().SetState(False)
        else:
            for actor in self.actor_meshes:
                self.p.remove_actor(actor)

    def plotNormals(self, value):
        if value:
            for mesh in self.meshes:
                vecLength = np.amax(pdist(mesh.points))
                normals = vecLength * mesh.point_normals
                self.actor_normals.append(self.p.add_arrows(mesh.points, normals,
                                          mag=0.1, color='blue'))
        else:
            for actor in self.actor_normals:
                self.p.remove_actor(actor)

    def plotExtNormals(self, value):
        if value:
            for idn, normals in enumerate(self.extNormals_List):
                normals = pv.pyvista_ndarray(normals)
                vecLength = np.amax(pdist(self.meshes[idn].points))
                normals = vecLength * (normals / np.linalg.norm(normals, axis=1)[:, np.newaxis])
                if self.extNormals_coords is None:
                    areZero = np.where((self.meshes[idn].point_normals == (0, 0, 0)).all(axis=1))
                    normals[areZero] = np.array((0, 0, 0))
                    self.actor_extNormals.append(self.p.add_arrows(self.meshes[idn].points, normals,
                                                                   mag=0.1, color='red'))
                else:
                    extNormals_coords = pv.pyvista_ndarray(self.extNormals_coords)
                    self.actor_extNormals.append(self.p.add_arrows(extNormals_coords, normals,
                                                                   mag=0.1, color='red'))
        else:
            for actor in self.actor_extNormals:
                self.p.remove_actor(actor)

    def plotPoints(self, value):
        if value:
            for mesh in self.meshes:
                self.actor_points.append(self.p.add_points(mesh, color='black', point_size=10,
                                                           render_points_as_spheres=True))
            self.buttonMesh.GetRepresentation().SetState(False)
        else:
            for actor in self.actor_points:
                self.p.remove_actor(actor)

    def plotOriPoints(self, value):
        if value:
            for cloud in self.clouds:
                self.actor_ori_points.append(self.p.add_points(cloud, color='white', point_size=10,
                                                               render_points_as_spheres=True))
            if self.buttonMesh.GetRepresentation().GetState():
                for actor in self.actor_meshes:
                    self.p.remove_actor(actor)
                self.buttonMesh.GetRepresentation().SetState(False)
        else:
            for actor in self.actor_ori_points:
                self.p.remove_actor(actor)

    def initializePlot(self):
        self.app.exec_()
        
# The following functions are used to create automatically a "gui thread" to avoid the 'exec_' loop 
# from freezing the main thread

def guiThread(classObj, methodName, *args, **kwargs):
    '''
    Create a new process to prevent the exec_ loop of the GUI from blocking the main thread. In order to work,
    the GUI classes must be instantiated withing the process.
    '''
    proc = Process(target=instantiateClass, args=(classObj, methodName, *args,), kwargs=kwargs)
    proc.start()
    
def instantiateClass(classObj, methodName, *args, **kwargs):
    '''
    Create an instance of any class and call a visualization method (or any other method).
        - classObj (class): Class to be instantiated
        - methodName (string): Method from the class to be called after the instantiation
        - *args (list): extra argument needed to call the class method
        - **kwargs (dict): arguments to be passed to the contructor of the class
    '''
    try:
        instance = classObj(**kwargs)
        runMethod(instance, methodName, *args)
    except:
        print('Cannot create instance of class')
    
def runMethod(instance, methodName, *args):
    '''
    Execute a method from an instantiated class:
        - instance (obj): object instantiated from a given class
        - methodName (string): method belonging to instance to be called
        - *args (list): extra arguments needed by the method
    '''
    try:
        method = getattr(instance, methodName)
        method(*args)
    except AttributeError:
        print(methodName + ' is not a member the class')



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     Estrella Fernandez Gimenez (me.fernandez@cnb.csic.es)
# *
# *  BCU, Centro Nacional de Biotecnologia, CSIC
# *
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from os import path
import numpy as np
from pyworkflow import BETA
from pyworkflow.protocol.params import PointerParam, FloatParam, BooleanParam, IntParam
import pwem.convert.transformations as tfs
from pwem.protocols import EMProtocol
from tomo.objects import SubTomogram, Coordinate3D
from tomo.protocols import ProtTomoBase
from tomo.utils import normalFromMatrix
import tomo.constants as const
from ..utils import delaunayTriangulation, computeNormals


class XmippProtFilterbyNormal(EMProtocol, ProtTomoBase):
    """ This protocol takes surfaces or ROIs (SetOfMeshes) and a SetOfSubtomograms or SetOfCoordinates3D with
    transformation matrix and filters them by different criteria related with the normal direction."""

    _label = 'filter by normal'
    _devStatus = BETA

    def __init__(self, **args):
        EMProtocol.__init__(self, **args)

    # --------------------------- DEFINE param functions ------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('input', PointerParam, pointerClass="SetOfSubTomograms, SetOfCoordinates3D",
                      label='Subtomograms/Coordinates', help='SetOfSubtomograms to filter.')
        form.addParam('inputMeshes', PointerParam, label="Vesicles", pointerClass='SetOfMeshes',
                      help='Select the vesicles in which the subtomograms/coordinates are.')
        form.addParam('tilt', BooleanParam, default=False,
                      label='Filter by tilt angle',
                      help='Remove items depending on their tilt angle.')
        form.addParam('maxtilt', IntParam, default=150, label='Maximum allowed tilt', condition='tilt',
                      help='Remove the items that have a tilt angle bigger than the one specified in here, '
                           'considering tilt angle between 0 and 180 degrees.')
        form.addParam('mintilt', IntParam, default=30, label='Minimum allowed tilt', condition='tilt',
                      help='Remove the items that have a tilt angle smaller than the one specified in here, '
                           'considering tilt angle between 0 and 180 degrees.')
        form.addParam('normalDir', BooleanParam, default=True,
                      label='Filter by normal',
                      help='Remove the items that have a normal direction not equal to the normal direction of '
                           'the vesicle in the coordinate of the particle.')
        form.addParam('tol', FloatParam, default=5, label='Tolerance in degrees',
                      condition='normalDir',
                      help='Tolerance (in degrees) when comparing between subtomogram/coordinate and mesh normal '
                           'directions.')

        # form.addParam('topBottom', BooleanParam, default=True,
        #               label='Remove subtomograms in the top and bottom of the vesicle',
        #               help='Remove the subtomograms that have been picked from the top and bottom parts because they '
        #                    'had a different view.')
        # form.addParam('mwDir', BooleanParam, default=True,
        #               label='Remove subtomograms in the missing wedge direction',
        #               help='Remove the subtomograms that are in the missing wedge direction because they are highly '
        #                    'affected by the missing wedge.')
        # form.addParam('mwDir', EnumParam, default=True, label='Missing wedge direction',
        #               help='Missing wedge direction of the tomograms.')

    # --------------------------- INSERT steps functions --------------------------------------------
    def _insertAllSteps(self):
        self._insertFunctionStep('computeNormalStep')
        self._insertFunctionStep('createOutputStep')

    # --------------------------- STEPS functions -------------------------------
    def computeNormalStep(self):
        inSet = self.input.get()
        tiltBool = self.tilt.get()
        normalBool = self.normalDir.get()
        inMeshes = self.inputMeshes.get()
        if normalBool:
            tol = self.tol.get() * np.pi / 180

            groupIdList = []
            tomoNameList = []
            for meshPoint in inMeshes:
                groupId = meshPoint.getGroupId()
                if groupId not in groupIdList:
                    groupIdList.append(groupId)
                    tomoNameList.append(str(meshPoint.getVolumeName()))

            meshDict = {key: {'tomoName': [tomoName], 'points': [], 'normals': []}
                        for key, tomoName in zip(groupIdList, tomoNameList)}

            for meshPoint in inMeshes.iterCoordinates():
                if meshDict[meshPoint.getGroupId()]["tomoName"][0] == meshPoint.getVolumeName():
                    meshDict[meshPoint.getGroupId()]["points"].append(meshPoint.getPosition(const.SCIPION))

            for i in meshDict:
                meshDict[i]["normals"] = self._getNormalVesicleList(np.asarray(meshDict[i]["points"]))

        if self._getInputisSubtomo(inSet.getFirstItem()):
            self.outSet = self._createSetOfSubTomograms()
        else:
            self.outSet = self._createSetOfCoordinates3D(inMeshes.getPrecedents())
        self.outSet.copyInfo(inSet)

        if tiltBool:
            if self._getInputisSubtomo(inSet.getFirstItem()):
                for item in inSet:
                    tilt = self._getTilt(item)
                    if self.maxtilt.get() > tilt > self.mintilt.get():
                        if normalBool:
                            self._filterByNormal(item, tol, meshDict)
                        else:
                            self.outSet.append(item)
            else:
                for item in inSet.iterCoordinates(volume=None):
                    tilt = self._getTilt(item)
                    if self.maxtilt.get() > tilt > self.mintilt.get():
                        if normalBool:
                            self._filterByNormal(item, tol, meshDict)
                        else:
                            self.outSet.append(item)

        if normalBool and not tiltBool:
            if self._getInputisSubtomo(inSet.getFirstItem()):
                for item in inSet:
                    self._filterByNormal(item, tol, meshDict)
            else:
                for item in inSet.iterCoordinates(volume=None):
                    self._filterByNormal(item, tol, meshDict)

    def createOutputStep(self):
        self._defineOutputs(outputset=self.outSet)
        self._defineSourceRelation(self.input.get(), self.outSet)

    # --------------------------- INFO functions --------------------------------
    def _validate(self):
        validateMsgs = []
        if not self.normalDir.get() and not self.tilt.get():
            validateMsgs.append('Some filter should be switched to "Yes"')
        if not self.input.get().getFirstItem().hasTransform():
            validateMsgs.append('Input coordinates/subtomograms should have transform matrix.')
        return validateMsgs

    def _summary(self):
        summary = []
        if not self.isFinished():
            summary.append("Output subtomograms/coordinates not ready yet.")
        else:
            if self.tilt:
                summary.append("Remove input items by tilt angle (max allowed tilt: %d, min allowed tilt: %d)" %
                               (self.maxtilt.get(), self.mintilt.get()))
            if self.normalDir:
                summary.append("Remove input items by normal direction (tolerance of %0.2f degrees)" %
                               self.tol.get())
        return summary

    def _methods(self):
        methods = []
        if not self.isFinished():
            methods.append("Output subtomograms/coordinates not ready yet.")
        else:
            if self.tilt:
                methods.append("Input items with tilt angle bigger than %d or smaller than %d have been removed." %
                               (self.maxtilt.get(), self.mintilt.get()))
            if self.normalDir:
                methods.append("Input items that are not perpendicular to the membrane have been removed.")

            # if self.topBottom:
            #     methods.append("Particles in the top and bottom parts of the vesicles have been removed.")
            # if self.mwDir:
            #     methods.append("Particles in the missing wedge direction have been removed.")
        return methods

    # --------------------------- UTILS functions --------------------------------------------
    def _getInputisSubtomo(self, item):
        if isinstance(item, SubTomogram):
            return True
        else:
            return False

    def _getVesicleId(self, item):
        if isinstance(item, SubTomogram):
            c = item.getCoordinate3D()
        else:
            c = item
        if c.hasGroupId():
            vesicleId = c.getGroupId()
        else:  # For now it works with several vesicles in the same tomo just for input items with groupId
            vesicleId = 1
        return vesicleId

    def _getNormalVesicleList(self, points):
        triangulation = delaunayTriangulation(points)
        normalsList = computeNormals(triangulation, associateCoords=True)
        return normalsList

    def _getNormalVesicle(self, normalsList, item):
        if self._getInputisSubtomo(item):
            normSubtomo = normalFromMatrix(item.getTransform().getMatrix())
            coord = item.getCoordinate3D()
        else:
            normSubtomo = normalFromMatrix(item.getMatrix())
            coord = item
        coors = np.asarray([coord.getX(const.SCIPION),
                            coord.getY(const.SCIPION),
                            coord.getZ(const.SCIPION)])
        points, normals = zip(*normalsList)
        points = np.asarray(points)
        idx = np.argmin(np.sum((points - coors) ** 2, axis=1))
        return normSubtomo, normals[idx]

    def _getTilt(self, item):
        if self._getInputisSubtomo(item):
            _, tilt, _ = tfs.euler_from_matrix(item.getTransform().getMatrix(), axes='szyz')
        else:
            _, tilt, _ = tfs.euler_from_matrix(item.getMatrix(), axes='szyz')
        tilt = -np.rad2deg(tilt)
        return tilt

    def _filterByNormal(self, item, tol, meshDict):
        meshfromDict = meshDict[self._getVesicleId(item)]
        if meshfromDict["tomoName"][0] == path.basename(item.getVolName()):
            normSubtomo, normVesicle = self._getNormalVesicle(meshfromDict["normals"], item)
            if abs(normSubtomo[0] - normVesicle[0]) < tol and abs(normSubtomo[1] - normVesicle[1]) < tol and \
                    abs(normSubtomo[2] - normVesicle[2]) < tol:
                self.outSet.append(item)



# **************************************************************************
# *
# * Authors:    Carlos Oscar Sorzano (coss@cnb.csic.es)
# *             Tomas Majtner (tmajtner@cnb.csic.es)  -- streaming version
# *             David Herreros Calero (dherreros@cnb.csic.es) -- Tomo version
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'coss@cnb.csic.es'
# *
# **************************************************************************
"""
Consensus picking protocol
"""

import os

from math import sqrt
import numpy as np

from pyworkflow import BETA
from pyworkflow.object import Set, Pointer
import pyworkflow.protocol.params as params
from pyworkflow.protocol.constants import *
from pyworkflow.utils import getFiles, removeBaseExt, moveFile, removeExt

from pwem.convert.transformations import quaternion_from_matrix, weighted_tensor,\
                                         mean_quaternion, quaternion_matrix

from tomo.protocols import ProtTomoPicking
from tomo.objects import SetOfCoordinates3D, Coordinate3D
import tomo.constants as const

PICK_MODE_LARGER = 0
PICK_MODE_EQUAL = 1

class ProtTomoConsensusPicking(ProtTomoPicking):
    """
    Protocol to estimate the agreement between different particle picking
    algorithms. The protocol takes several Sets of Coordinates calculated
    by different programs and/or different parameter settings. Let's say:
    we consider N independent pickings. Then, a coordinate is considered
    to be a correct particle if M pickers have selected the same particle
    (within a radius in pixels specified in the form).

    If you want to be very strict, then set M=N; that is, a coordinate
    represents a particle if it has been selected by all particles (this
    is the default behaviour). Then you may relax this condition by setting
    M=N-1, N-2, ...

    If you want to be very flexible, set M=1, in this way it suffices that
    1 picker has selected the coordinate to be considered as a particle. Note
    that in this way, the cleaning of the dataset has to be performed by other
    means (screen particles, 2D and 3D classification, ...).
    """

    _label = 'picking consensus'
    _devStatus = BETA
    outputName = 'consensusCoordinates'
    FN_PREFIX = 'consensusCoords_'

    def __init__(self, **args):
        ProtTomoPicking.__init__(self, **args)
        self.stepsExecutionMode = STEPS_SERIAL

    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('inputCoordinates', params.MultiPointerParam,
                      pointerClass='SetOfCoordinates3D',
                      label="Input 3D coordinates", important=True,
                      help='Select the set of 3D coordinates to compare')
        form.addParam('consensusRadius', params.IntParam, default=10,
                      label="Radius",
                      help="All coordinates within this radius (in pixels) "
                           "are presumed to correspond to the same particle")
        form.addParam('consensus', params.IntParam, default=-1,
                      label="Consensus",
                      help="This parameter can take values from 1 to total number of inputs (being "
                           "-1 a special case).\n"
                           "*Set to -1* to indicate that it needs to be selected "
                           "by all algorithms: *AND* operation.\n"
                           "*Set to 1* to indicate that it suffices that only "
                           "1 algorithm selects the particle: *OR* operation.\n"
                           "Any other value will determine how many times need a particle"
                           " to be selected to be considered as a consensus particle.")
        form.addParam('mode', params.EnumParam, label='Consensus mode',
                      choices=['>=', '='], default=PICK_MODE_LARGER,
                      expertLevel=LEVEL_ADVANCED,
                      display=params.EnumParam.DISPLAY_HLIST,
                      help='If the number of votes to progress to the output '
                           'must be either (=) strictly speaking equals to '
                           'the consensus number or (>=) at least equals.')

        # FIXME: It's not using more than one since
        #         self.stepsExecutionMode = STEPS_SERIAL
        # form.addParallelSection(threads=4, mpi=0)

#--------------------------- INSERT steps functions ---------------------------
    def _insertAllSteps(self):
        self.checkedTomos = set()   # those tomos ready to be processed (tomoId)
        self.processedTomos = set() # those tomos already processed (tomoId)
        self.sampligRates = []
        coorSteps = self.insertNewCoorsSteps([])
        self._insertFunctionStep('createOutputStep',
                                 prerequisites=coorSteps, wait=True)

    def createOutputStep(self):
        pass

    def _getFirstJoinStepName(self):
        # This function will be used for streaming, to check which is
        # the first function that need to wait for all mics
        # to have completed, this can be overriden in subclasses
        # (e.g., in Xmipp 'sortPSDStep')
        return 'createOutputStep'

    def _getFirstJoinStep(self):
        for s in self._steps:
            if s.funcName == self._getFirstJoinStepName():
                return s
        return None

    def insertNewCoorsSteps(self, tomos):
        deps = []
        for tomo in tomos:
            stepId = self._insertFunctionStep("calculateConsensusStep",
                                              tomo.getObjId(),
                                              tomo.getFileName(),
                                              prerequisites=[])
            deps.append(stepId)
        return deps

    def _stepsCheck(self):
        self._checkNewInput()
        self._checkNewOutput()

    def _checkNewInput(self):
        # If continue from an stopped run, don't repeat what is done
        if not self.checkedTomos:
            for fn in getFiles(self._getExtraPath()):
                fn = removeBaseExt(fn)
                if fn.startswith(self.FN_PREFIX):
                    self.checkedTomos.update([self.getTomoId(fn)])
                    self.processedTomos.update([self.getTomoId(fn)])

        streamClosed = []
        readyTomos = None
        allTomos = set()
        for coordSet in self.inputCoordinates:
            currentPickTomos, isSetClosed = getReadyTomos(coordSet.get())
            streamClosed.append(isSetClosed)
            if not readyTomos:  # first time
                readyTomos = currentPickTomos
            else:  # available mics are those ready for all pickers
                readyTomos.intersection_update(currentPickTomos)
            allTomos = allTomos.union(currentPickTomos)

        self.streamClosed = all(streamClosed)
        if self.streamClosed:
            # for non streaming do all and in the last iteration of streaming do the rest
            newTomoIds = allTomos.difference(self.checkedTomos)
        else:  # for streaming processing, only go for the ready mics in all pickers
            newTomoIds = readyTomos.difference(self.checkedTomos)

        if newTomoIds:
            self.checkedTomos.update(newTomoIds)

            inTomos = self.getMainInput().getPrecedents()
            newTomos = [inTomos[tomoId].clone() for tomoId in newTomoIds]

            fDeps = self.insertNewCoorsSteps(newTomos)
            outputStep = self._getFirstJoinStep()
            if outputStep is not None:
                outputStep.addPrerequisites(*fDeps)
            self.updateSteps()

    def _checkNewOutput(self):
        if getattr(self, 'finished', False):
            return
        self.finished = self.streamClosed and self.checkedTomos == self.processedTomos
        streamMode = Set.STREAM_CLOSED if self.finished else Set.STREAM_OPEN

        newFiles = getFiles(self._getTmpPath())
        if newFiles or self.finished:  # when finished to close the output set
            outSet = self._loadOutputSet(SetOfCoordinates3D, 'coordinates.sqlite')

            for fnTmp in newFiles:
                if '_vIds' in fnTmp or '_trMats' in fnTmp:
                    continue
                coords = np.loadtxt(fnTmp)
                vsIds = np.loadtxt(removeExt(fnTmp) + '_vIds.txt')
                with open(removeExt(fnTmp) + '_trMats.txt') as outFile:
                    shape_mat = outFile.readline()
                    shape_mat = eval('(' + shape_mat.split('(')[1])
                    trMats = np.loadtxt(removeExt(fnTmp) + '_trMats.txt').reshape(shape_mat)
                moveFile(fnTmp, self._getExtraPath())
                moveFile(removeExt(fnTmp) + '_vIds.txt', self._getExtraPath())
                for idv in np.unique(vsIds):
                    vesicle_coords = coords[np.where(vsIds == idv)]
                    vesicle_tr = trMats[np.where(vsIds == idv)]
                    if vesicle_coords.size == 3:  # special case with only one coordinate
                        vesicle_coords = [vesicle_coords]
                    for idx, coord in enumerate(vesicle_coords):
                        newCoord = Coordinate3D()
                        tomograms = self.getMainInput().getPrecedents()
                        newCoord.setVolume(tomograms[self.getTomoId(fnTmp)])
                        newCoord.setPosition(coord[0], coord[1], coord[2], const.SCIPION)
                        newCoord.setGroupId(idv)
                        matrix = vesicle_tr[idx]
                        if isinstance(self.inputCoordinates, list):
                            newCoord.setMatrix(matrix)
                        else:
                            newCoord.setMatrix(matrix)
                        outSet.append(newCoord)

            firstTime = not self.hasAttribute(self.outputName)
            self._updateOutputSet(self.outputName, outSet, streamMode)
            if firstTime:
                self.defineRelations(outSet)
            outSet.close()

        if self.finished:  # Unlock createOutputStep if finished all jobs
            outputStep = self._getFirstJoinStep()
            if outputStep and outputStep.isWaiting():
                outputStep.setStatus(STATUS_NEW)

    def defineRelations(self, outputSet):
        for inCorrds in self.inputCoordinates:
            self._defineTransformRelation(inCorrds, outputSet)

    def _loadOutputSet(self, SetClass, baseName):
        setFile = self._getPath(baseName)
        if os.path.exists(setFile):
            outputSet = SetClass(filename=setFile)
            outputSet.loadAllProperties()
            outputSet.enableAppend()
        else:
            outputSet = SetClass(filename=setFile)
            outputSet.setStreamState(outputSet.STREAM_OPEN)

        #inMicsPointer = Pointer(self.getMapper().getParent(
        #                                    self.getMainInput().getMicrographs()),
        #                                    extended='outputMicrographs')
        outputSet.setBoxSize(self.getMainInput().getBoxSize())
        outputSet.setSamplingRate(self.getMainInput().getSamplingRate())
        inTomosPointer = Pointer(self.getMainInput().getPrecedents())
        outputSet.setPrecedents(inTomosPointer)

        return outputSet

    def calculateConsensusStep(self, tomoId, tomoName):

        print("Consensus calculation for Tomogram %d: '%s'"
              % (tomoId, tomoName))

        # Take the sampling rates just once
        if not self.sampligRates:
            for coordinates in self.inputCoordinates:
                tomograms = coordinates.get().getPrecedents()
                self.sampligRates.append(tomograms.getSamplingRate())

        # Get all coordinates for this tomogram
        coords = []
        vesicles = []
        transformations = []
        for idx, coordinates in enumerate(self.inputCoordinates):
            coordArray = np.asarray([x.getPosition(const.SCIPION) for x in
                                     coordinates.get().iterCoordinates(tomoId)],
                                     dtype=float)
            vIds = np.asarray([x.getGroupId() for x in
                               coordinates.get().iterCoordinates(tomoId)])
            trMat = np.asarray([x.getMatrix() for x in
                                coordinates.get().iterCoordinates(tomoId)])
            coordArray *= float(self.sampligRates[idx]) / float(self.sampligRates[0])
            coords.append(np.asarray(coordArray, dtype=int))
            vesicles.append(np.asarray(vIds, dtype=int))
            transformations.append(trMat)

        consensusWorker(coords, vesicles, transformations, self.consensus.get(),
                        self.consensusRadius.get(),
                        self._getTmpPath('%s%s.txt' % (self.FN_PREFIX, tomoId)),
                        self._getExtraPath('jaccard.txt'), self.mode.get())

        self.processedTomos.update([tomoId])

    def _validate(self):

        errors = []

        # Only for Scipion 2.0, next versions should have the default
        # PointerList validation and this can be removed
        if len(self.inputCoordinates) == 0:
                errors.append('inputCoordinates cannot be EMPTY.')
        # Consider empty pointers:
        else:
            for pointer in self.inputCoordinates:
                obj = pointer.get()
                if obj is None:
                    errors.append('%s is empty.' % obj)

        return errors

    def _summary(self):
        message = []
        for i, coordinates in enumerate(self.inputCoordinates):
            protocol = self.getMapper().getParent(coordinates.get())
            message.append("Method %d %s" % (i + 1, protocol.getClassLabel()))
        message.append("Radius = %d" % self.consensusRadius)
        message.append("Consensus = %d" % self.consensus)
        return message

    def _methods(self):
        return []

    def getMainInput(self):
        return self.inputCoordinates[0].get()

    @classmethod
    def getTomoId(self, fn):
        return int(removeBaseExt(fn).lstrip(self.FN_PREFIX))


def consensusWorker(coords, vesicles, trMats, consensus, consensusRadius, posFn, jaccFn=None,
                    mode=PICK_MODE_LARGER):
    """ Worker for calculate the consensus of N picking algorithms of
          M_n coordinates each one.

        coords: Array of N numpy arrays of M_n coordinates each one.
        consensus: Minimum number of votes to get a consensus coordinate
        consensusRadius: Tolerance to see two coordinates as the same (in pixels)
        posFn: Where to write the consensus coordinates
        jaccFN: Where to write the Jaccard index per micrograph
    """
    if len(coords) == 1:  # self consensus (remove duplicates)
        N0 = 0
        firstInput = 0
    else:  # in regular consensus all first coords are directly added to allCoords
        N0 = coords[0].shape[0]
        firstInput = 1


    # initializing arrays
    Ninputs = len(coords)
    Ncoords = sum([x.shape[0] for x in coords])
    allCoords = np.zeros([Ncoords, 3])
    allVesicles = np.zeros([Ncoords])
    allQuaternions = np.zeros([Ncoords, 4])
    votes = np.zeros(Ncoords)

    # Convert transformation matrices (rotations) to quaternions
    quaternions = []
    for n in range(len(trMats)):
        tomoMats = trMats[n]
        quatertions_tomo = np.asarray([quaternion_from_matrix(tomoMats[idq]) for idq in range(len(tomoMats))])
        quaternions.append(quatertions_tomo)

    inAllMicrographs = consensus <= 0 or consensus >= Ninputs

    # if nothing in the first and it should be in all, nothing to do
    if (not all([coords[idx].shape[0] for idx in range(Ninputs)])
            and inAllMicrographs):
        print("Returning from worker: doing AND consensus and, at least, one "
              "picker is empty for this micrograph (%s)." % posFn)
        return

    # Add all the first coordinates to 'allCoords' and 'votes' lists
    if N0 > 0:
        allCoords[0:N0, :] = coords[0]
        allVesicles[0:N0] = vesicles[0]
        allQuaternions[0:N0] = quaternions[0]
        votes[0:N0] = 1

    # Add the rest of coordinates to 'allCoords' and 'votes' lists
    Ncurrent = N0
    for n in range(firstInput, Ninputs):
        for idv in np.unique(vesicles[n]):
            coords_vesicle = coords[n][np.where(vesicles[n] == idv)]
            q_vesicle = quaternions[n][np.where(vesicles[n] == idv)]
            for coord, q in zip(coords_vesicle, q_vesicle):
                if Ncurrent > 0:
                    dist = np.sum((coord - allCoords[0:Ncurrent]) ** 2, axis=1)
                    imin = np.argmin(dist)
                    if sqrt(dist[imin]) < consensusRadius:
                        newCoord = (votes[imin] * allCoords[imin,] + coord) / (
                                    votes[imin] + 1)
                        allCoords[imin,] = newCoord
                        tuple_q = (allQuaternions[imin,], q)
                        T = weighted_tensor(tuple_q)
                        q_bar, _ = mean_quaternion(T)
                        allQuaternions[imin,] = q_bar
                        allVesicles[imin] = idv
                        votes[imin] += 1
                    else:
                        allCoords[Ncurrent, :] = coord
                        allVesicles[Ncurrent] = idv
                        allQuaternions[Ncurrent] = q
                        votes[Ncurrent] = 1
                        Ncurrent += 1
                else:
                    allCoords[Ncurrent, :] = coord
                    allVesicles[Ncurrent] = idv
                    allQuaternions[Ncurrent] = q
                    votes[Ncurrent] = 1
                    Ncurrent += 1

    # Select those in the consensus
    if consensus <= 0 or consensus > Ninputs:
        consensus = Ninputs
    elif not isinstance(consensus, int):
        consensus = consensus.get()

    if mode==PICK_MODE_LARGER:
        consensusCoords = allCoords[votes >= consensus, :]
        qConsensus = allQuaternions[votes >= consensus]
        vesiclesConsensus = allVesicles[votes >= consensus]
    else:
        consensusCoords = allCoords[votes == consensus, :]
        qConsensus = allQuaternions[votes == consensus]
        vesiclesConsensus = allVesicles[votes == consensus]

    # Convert consensus quaternions back to consensus transformations
    trConsensus = np.zeros([len(qConsensus), 4, 4])
    for idq, q in enumerate(qConsensus):
        tr = quaternion_matrix(q)
        trConsensus[idq,] = tr

    try:
        if jaccFn:
            jaccardIdx = float(len(consensusCoords)) / (
                    float(len(allCoords)) / Ninputs)
            # COSS: Possible problem with concurrent writes
            with open(jaccFn, "a") as fhJaccard:
                fhJaccard.write("%s \t %f\n" % (posFn, jaccardIdx))
    except Exception as exc:
        print("Some error occurred during Jaccard index calculation or "
              "writing it's file. Maybe a concurrence issue:\n%s" % exc)
    # Write the consensus file only if there
    # are some coordinates (size > 0)
    if consensusCoords.size:
        np.savetxt(posFn, consensusCoords)
        np.savetxt(removeExt(posFn) + '_vIds.txt', vesiclesConsensus)
        with open(removeExt(posFn) + '_trMats.txt', 'w') as outfile:
            outfile.write('# Array shape: {0}\n'.format(trConsensus.shape))
            for tr_slice in trConsensus:
                np.savetxt(outfile, tr_slice, fmt='%-7.2f')
                outfile.write('# New Transformation Matrix\n')


def getReadyTomos(coordSet):
    coorSet = SetOfCoordinates3D(filename=coordSet.getFileName())
    coorSet.loadAllProperties()
    setClosed = coorSet.isStreamClosed()
    coorSet.close()
    currentPickTomos = {tomoAgg["_volId"] for tomoAgg in
                        coordSet.aggregate(["MAX"], "_volId", ["_volId"])}
    return currentPickTomos, setClosed



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from pyworkflow.tests import DataSet

DataSet(name='tomo-em', folder='tomo-em',
        files={
               'tomo1': 'overview_wbp.em',
               'tomo2': 'overview_wbp2.em',
               'tomo3': 'tomo_8_mn.mrc',
               'subtomo': 'basename.hdf',
               'eman_coordinates': 'coordinates3Deman2',
               'etomo': 'tutorialData',
               'empiar': 'EMPIAR-10164'
        })

DataSet(name='reliontomo', folder='reliontomo',
        files={
               'tomo1': 'defocus_m2_tomo_10_bin1_WBP_CatBinned1.mrc',
               'tomo2': 'defocus_m2_tomo_12_bin1_WBP_CatBinned1.mrc',
        })



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import pyvistaqt as pvqt
import pyvista as pv
from pyvista.utilities import generate_plane
import vtk
import numpy as np

class VtkPlot(object):
    '''
    Class to visualize VTK files
    Input paramters:
         - vti_file (Path (Str) - Optional): File containing a Volume (Volume VTK Object)
         - graph_file (Path (Str) - Optional): File containing a Graph (PolyData VTK Object)
         - net_file (Path (Str) - Optional): File containing a Net (PolyData VTK Object)
         - peaks_file (Path (Str) - Optional): File containing Peaks - Coordinates (PolyData VTK Object)
         - surf_file (Path (Str) - Optional): File containing a Surface (PolyData VTK Object)
    Usage:
         import VtkPlot
         plt = VtkPlot(vti_file=path_vti, graph_file=path_graph, net_file=path_net, peaks_file=path_peaks)
         plt.initializePlot()
    '''

    def __init__(self, vti_file=None, graph_file=None, net_file=None, peaks_file=None, surf_file=None):
        self.vti = pv.read(vti_file) if vti_file is not None else None
        self.graph = pv.read(graph_file) if graph_file is not None else None
        self.net = pv.read(net_file) if net_file is not None else None
        self.peaks = pv.read(peaks_file) if peaks_file is not None else None
        self.surf = pv.read(surf_file) if surf_file is not None else None

        self.vti_actor = None
        self.graph_actor = None
        self.net_actor = None
        self.net_slice_actor = None
        self.peaks_actor = None
        self.vectors_actor = None
        self.surf_actor = None

        self.plane_widgets = []
        self.cut_origin = None
        self.cut_normal = None

        self.plt = pvqt.BackgroundPlotter(title='Scipion tomoviz viewer', window_size=(1200, 800))
        self.plt.main_menu.clear()

        pos = 0.

        def function_builder(vtk_obj, vtk_actor, vtk_button, vtk_actor_name, prop):
            '''
            Function to create in a automatic manner the contents of the different custom menus in
            pyvistaqt
            '''
            def function():
                self.plt.remove_actor(vtk_actor)
                setattr(self, vtk_actor_name,
                        self.plt.add_mesh(vtk_obj, show_scalar_bar=False, colormap="cool", scalars=prop))
                vtk_button.GetRepresentation().SetState(True)

            return function

        if self.vti:
            pos += 45.
            self.buttonVti = self.plt.add_checkbox_button_widget(callback=self.plotVti, position=(pos, 10.))
            self.plt.add_text('Tomogram', position=(pos, 65.), font_size=12)
            pos += 170.
            self.buttonSliceVti = self.plt.add_checkbox_button_widget(callback=self.toogleTomoSlice, position=(pos, 10.))
            self.plt.add_text('Tomo Slice', position=(pos, 65.), font_size=12)

        if self.graph:
            pos += 170. if pos != 0 else 45.
            self.plt.add_text('Graph', position=(pos, 65.), font_size=12)
            self.buttonGraph = self.plt.add_checkbox_button_widget(callback=self.plotGraph, position=(pos, 10.))

            # Color By Menu
            self.callbacks_graph = {}
            graph_properties = np.sort(self.graph.array_names)
            graph_menu = self.plt.main_menu.addMenu('Color Graph By')
            for prop in graph_properties:
                if 'normal' not in prop:
                    self.callbacks_graph[prop] = function_builder(self.graph, self.graph_actor,
                                                                  self.buttonGraph, 'graph_actor', prop)
                    graph_menu.addAction(prop, self.callbacks_graph[prop])

        if self.net:
            pos += 170. if pos != 0 else 45.
            self.plt.add_text('Net', position=(pos, 65.), font_size=12)
            self.buttonNet = self.plt.add_checkbox_button_widget(callback=self.plotNet, position=(pos, 10.))
            pos += 170.
            self.buttonSliceNet = self.plt.add_checkbox_button_widget(callback=self.toogleNetSlice, position=(pos, 10.))
            self.plt.add_text('Net Slice', position=(pos, 65.), font_size=12)

            # Color By Menu
            self.callbacks_net = {}
            net_properties = np.sort(self.net.array_names)
            net_menu = self.plt.main_menu.addMenu('Color Net By')
            for prop in net_properties:
                if 'normal' not in prop:
                    self.callbacks_net[prop] = function_builder(self.net, self.net_actor,
                                                                self.buttonNet, 'net_actor', prop)
                    net_menu.addAction(prop, self.callbacks_net[prop])

        if self.peaks:
            self.peaks.set_active_vectors('smb_normal')
            pos += 170. if pos != 0 else 45.
            self.plt.add_text('Peaks', position=(pos, 65.), font_size=12)
            self.buttonPeaks = self.plt.add_checkbox_button_widget(callback=self.plotPeaks, position=(pos, 10.))
            pos += 170.
            self.buttonVectors = self.plt.add_checkbox_button_widget(callback=self.plotVectors, position=(pos, 10.))
            self.plt.add_text('Directions', position=(pos, 65.), font_size=12)

        if self.surf:
            pos += 170. if pos != 0 else 45.
            self.plt.add_text('Surface', position=(pos, 65.), font_size=12)
            self.buttonPeaks = self.plt.add_checkbox_button_widget(callback=self.plotSurface, position=(pos, 10.))

    def plotVti(self, value):
        if value:
            self.vti_actor = self.plt.add_mesh_slice(self.vti, normal='z', cmap="bone", show_scalar_bar=False,
                                                     outline_translation=False, origin_translation=False)

            # In order to set our own callback to the widget, we need to turn off the current widget
            # And turning it on again with the custom function we want to use
            self.plt.plane_widgets[-1].Off()
            del self.plt.plane_widgets[-1]
            self.buttonSliceVti.GetRepresentation().SetState(True)
            self.toogleTomoSlice(True)

            self.plt.reset_camera()
        else:
            idx = self.plane_widgets.index(self.vti_actor)
            self.plt.plane_widgets[idx].Off()
            del self.plt.plane_widgets[idx]
            del self.plane_widgets[idx]
            self.plt.remove_actor(self.vti_actor)
            self.buttonSliceVti.GetRepresentation().SetState(False)
            self.vti_actor = None

            # We need to check if we have a Net Slice and remove if it exists
            if self.buttonSliceNet.GetRepresentation().GetState():
                self.plt.remove_actor(self.net_slice_actor)
                self.net_slice_actor = None
                self.buttonSliceNet.GetRepresentation().SetState(False)

            # We remove all the slice meshes and restore the plane widget variable of the viewer
            del self.plt.plane_sliced_meshes
            self.plane_widgets = []

    def toogleTomoSlice(self, value):
        if value:
            if self.buttonVti.GetRepresentation().GetState():
                if self.vti_actor in self.plane_widgets:
                    idx = self.plane_widgets.index(self.vti_actor)
                    self.plt.plane_widgets[idx].On()
                else:
                    # Since Tomo button will toggle this automatically, it is safe to append here the tomo
                    # actor to the currently active plane widgets
                    self.plane_widgets.append(self.vti_actor)

                    # Create all the vtkCutter objects that will be updated
                    alg_tomo, alg_net = vtk.vtkCutter(), vtk.vtkCutter()
                    alg_tomo.SetInputDataObject(self.vti)
                    alg_net.SetInputDataObject(self.net)

                    # This callback is called everytime the Tomo Slice (Cut) is updated
                    # by the user (move, rotation) or because it is newly created
                    def callback(normal, origin):
                        # Update cut normal and origin stored in the widget
                        self.cut_normal = normal
                        self.cut_origin = origin
                        # Create a new Plane based on new normal/origin
                        plane = generate_plane(normal, origin)
                        # Update Tomo Slice (Cut) position based on new plane
                        alg_tomo.SetCutFunction(plane)
                        alg_tomo.Update()
                        idx = self.plane_widgets.index(self.vti_actor)
                        plane_sliced_tomo = self.plt.plane_sliced_meshes[idx]
                        plane_sliced_tomo.shallow_copy(alg_tomo.GetOutput())
                        # This callback also takes into account the Net Plane
                        # In case it exists, it is update so it matches the current orientation of the
                        # Tomogram Slice
                        if self.buttonSliceNet.GetRepresentation().GetState():
                            alg_net.SetCutFunction(plane)
                            alg_net.Update()
                            idx = self.plane_widgets.index(self.net_slice_actor)
                            plane_sliced_net = self.plt.plane_sliced_meshes[idx]
                            plane_sliced_net.shallow_copy(alg_net.GetOutput())

                    self.plt.add_plane_widget(callback=callback, bounds=self.vti.bounds,
                                              factor=1.25, normal='z',
                                              origin_translation=False,
                                              outline_translation=False,
                                              origin=self.vti.center)
            else:
                self.buttonSliceVti.GetRepresentation().SetState(False)
        else:
            idx = self.plane_widgets.index(self.vti_actor)
            self.plt.plane_widgets[idx].Off()

    def plotGraph(self, value):
        if value:
            self.graph_actor = self.plt.add_mesh(self.graph, show_scalar_bar=False, colormap="cool")
        else:
            self.plt.remove_actor(self.graph_actor)
            self.graph_actor = None

    def plotNet(self, value):
        if value:
            self.net_actor = self.plt.add_mesh(self.net, show_scalar_bar=False, colormap="cool")
            if self.buttonSliceNet.GetRepresentation().GetState():
                self.buttonSliceNet.GetRepresentation().SetState(False)
                self.plt.remove_actor(self.net_slice_actor)
                self.net_slice_actor = None
        else:
            self.plt.remove_actor(self.net_actor)
            self.net_actor = None

    def toogleNetSlice(self, value):
        if value:
            if self.buttonVti.GetRepresentation().GetState():
                # We create a new Plane to perform a cut to Net
                alg = vtk.vtkCutter()
                alg.SetInputDataObject(self.net)
                plane_sliced_mesh = pv.wrap(alg.GetOutput())
                plane = generate_plane(self.cut_normal, self.cut_origin)
                alg.SetCutFunction(plane)
                alg.Update()
                plane_sliced_mesh.shallow_copy(alg.GetOutput())

                # Create the actor and add the plane to Pyvista plane_slice_meshes and plane_widgets to update
                # it in the future
                self.net_slice_actor = self.plt.add_mesh(plane_sliced_mesh, show_scalar_bar=False, colormap="cool")
                self.plt.plane_sliced_meshes.append(plane_sliced_mesh)
                self.plane_widgets.append(self.net_slice_actor)

                if self.buttonNet.GetRepresentation().GetState():
                    self.buttonNet.GetRepresentation().SetState(False)
                    self.plt.remove_actor(self.net_actor)
                    self.net_actor = None
            else:
                self.buttonSliceNet.GetRepresentation().SetState(False)
        else:
            self.plt.remove_actor(self.net_slice_actor)
            self.net_slice_actor = None
            # Since we only add the tomo before the Net Slice, we know that the last appended element to
            # the sliced meshes is going to be the Net Slice Mesh
            del self.plt.plane_sliced_meshes[-1]
            del self.plane_widgets[-1]


    def plotPeaks(self, value):
        mag = 0.02 * self.vti.dimensions[0]
        if value:
            self.peaks_actor = self.plt.add_points(self.peaks, color="orange", point_size=mag,
                                                   render_points_as_spheres=True, show_scalar_bar=False)
        else:
            self.plt.remove_actor(self.peaks_actor)
            self.peaks_actor = None

    def plotVectors(self, value):
        mag = 0.05 * self.vti.dimensions[0]
        if value:
            self.vectors_actor = self.plt.add_arrows(self.peaks.points, self.peaks.active_vectors, color="red",
                                                     mag=mag)
        else:
            self.plt.remove_actor(self.vectors_actor)
            self.vectors_actor = None

    def plotSurface(self, value):
        if value:
            self.surf_actor = self.plt.add_mesh(self.surf, opacity=0.6, color="white", show_scalar_bar=False)
        else:
            self.plt.remove_actor(self.surf_actor)
            self.surf_actor = None

    def initializePlot(self):
        self.plt.app.exec_()



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import pyvistaqt as pvqt
import pyvista as pv
from PyQt5.QtGui import QMovie
from pyvista.utilities import generate_plane
import pymeshfix as pm
import vtk
from PyQt5 import QtGui
from PyQt5.QtWidgets import QWidget, QLabel, QDesktopWidget
from PyQt5.QtCore import Qt, QObject, QThread, pyqtSignal, QRectF

import numpy as np
import matplotlib

from scipy.ndimage import zoom, gaussian_filter
# from scipy.spatial.distance import pdist

from skimage import measure
from skimage.morphology import binary_dilation, binary_erosion, ball

import pyworkflow.utils as pwutils

from pwem.emlib.image import ImageHandler

import tomoviz


class MrcPlot(object):
    '''
    Class to visualize MRC files
    Input paramters:
         - tomo_mrc (Path (Str) - Optional): File containing a Volume (in MRC format)
         - mask_mrc (Path (Str) - Optional): File containing a Mask (in MRC format)
         - points (Path (Str) - Optional): File containing a SetOfCoordinates3D (in TEXT format)
         - boxSize (Double - Optional): Box size associated to points
         - normals (Path (Str) - Optional): File containing a Set of Normals (in TEXT format)
         - binning (Float - Optional):  Binning factor to be applied to tomo_mrc and mask_mrc (Very useful to save time)
         - sigma (Float - Optional):  Gaussian Filter width
         - triangulation (Bool - Optional): Tells if the representation of the Tomogram is Voxel based on Triangle based
    Usage:
         import MRCPlot
         plt = MRCPlot(tomo_mrc=tomo_mrc, mask_mrc=mask_mrc, binning=2)
         plt.initializePlot()
    '''

    def __init__(self, tomo_mrc=None, mask_mrc=None, points=None, boxSize=None, normals=None,
                 binning=None, sigma=1., triangulation=False):
        if binning is None:
            if tomo_mrc is not None:
                self.binning = self.getBinning(tomo_mrc)
            elif mask_mrc is not None:
                self.binning = self.getBinning(mask_mrc)
            else:
                self.binning = 0
        else:
            self.binning = binning
        self.tomo = (tomo_mrc, triangulation, sigma)
        self.mask = mask_mrc
        self.points = np.loadtxt(points, delimiter=' ') if points is not None else None
        self.normals = np.loadtxt(normals, delimiter=' ') if normals is not None else None
        self.boxSize = boxSize / 2 ** self.binning if boxSize is not None else None
        self.save_basename = pwutils.removeBaseExt(tomo_mrc) if tomo_mrc is not None and points is not None else None

        # Get Pyvista Objects
        if isinstance(self.points, np.ndarray):
            self.points_ids = self.points[:, 3]
            self.group_ids = self.points[:, 4]
            self.points = np.column_stack([self.points[:, 1], self.points[:, 0], self.points[:, 2]])
            self.points /= 2 ** self.binning  # Binning Scaling
            self.pv_points = pv.PolyData(self.points)
            scalar_colors = np.zeros([self.points.shape[0], 4])
            cmap = matplotlib.cm.get_cmap('gist_rainbow_r')
            unique_ids = np.unique(self.group_ids)
            cmap_ids = np.linspace(0, 1, len(unique_ids))
            for group_id, cmap_id in zip(unique_ids, cmap_ids):
                idp = np.where(self.group_ids == group_id)
                scalar_colors[idp] = cmap(cmap_id)
            self.pv_points["colors"] = scalar_colors
        if isinstance(self.normals, np.ndarray):
            self.normals = np.column_stack([self.normals[:, 1], self.normals[:, 0], self.normals[:, 2]])
            # vecLength = np.amax(pdist(self.pv_points.points))
            self.normals /= np.linalg.norm(self.normals, axis=1)[:, np.newaxis]
            # self.normals *= vecLength
            self.pv_normals = pv.pyvista_ndarray(self.normals)

        self.tomo_actor = []
        self.tomo_slice_actor = None
        self.mask_actors = []
        self.points_actor = []
        self.normals_actor = None
        self.box_actor = {}

        self.first_reset = True

        self.plt = pvqt.BackgroundPlotter(title='Scipion tomoviz viewer')
        self.plt.main_menu.clear()
        plugin_path = os.path.dirname(tomoviz.__file__)
        self.plt.app.setWindowIcon(QtGui.QIcon(os.path.join(plugin_path, "icon_square.png")))
        self.loading_screen = LoadingScreen()

        pos = 0.

        if tomo_mrc is not None:
            pos += 45.
            self.buttonTomo = self.plt.add_checkbox_button_widget(callback=self.plotTomo, position=(pos, 10.))
            self.plt.add_text('Tomogram', position=(pos, 65.), font_size=12)
            pos += 170.
            self.buttonSliceTomo = self.plt.add_checkbox_button_widget(callback=self.toogleSlice, position=(pos, 10.))
            self.plt.add_text('Slice mode', position=(pos, 65.), font_size=12)

            self.plt.clear_events_for_key("Up")
            # Callback to move the slice with the up arrows in the keyboard
            def sliceUp():
                if self.tomo_slice_actor is not None:
                    alg_tomo = vtk.vtkCutter()
                    current_plane = vtk.vtkPlane()
                    alg_tomo.SetInputDataObject(self.pv_tomo_slice)
                    plane_sliced_tomo = self.plt.plane_sliced_meshes[-1]
                    plane_widget = self.plt.plane_widgets[-1]
                    plane_widget.GetPlane(current_plane)
                    origin = np.asarray(current_plane.GetOrigin())
                    normal = np.asarray(current_plane.GetNormal())
                    # Normalize normal
                    normal /= np.linalg.norm(normal)
                    # move plane one unit in the direction of the normal
                    origin += normal
                    # Create a new plane to update the position and perform the update
                    plane = generate_plane(normal, origin)
                    alg_tomo.SetCutFunction(plane)
                    alg_tomo.Update()
                    plane_sliced_tomo.shallow_copy(alg_tomo.GetOutput())
                    plane_widget.SetOrigin(origin)
                    plane_widget.SetNormal(normal)
                    plane_widget.UpdatePlacement()
            self.plt.add_key_event("Up", sliceUp)

            self.plt.clear_events_for_key("Down")
            # Callback to move the slice with the down arrows in the keyboard
            def sliceDown():
                if self.tomo_slice_actor is not None:
                    alg_tomo = vtk.vtkCutter()
                    current_plane = vtk.vtkPlane()
                    alg_tomo.SetInputDataObject(self.pv_tomo_slice)
                    plane_sliced_tomo = self.plt.plane_sliced_meshes[-1]
                    plane_widget = self.plt.plane_widgets[-1]
                    plane_widget.GetPlane(current_plane)
                    origin = np.asarray(current_plane.GetOrigin())
                    normal = np.asarray(current_plane.GetNormal())
                    # Normalize normal
                    normal /= np.linalg.norm(normal)
                    # move plane one unit in the direction of the normal
                    origin -= normal
                    # Create a new plane to update the position and perform the update
                    plane = generate_plane(normal, origin)
                    alg_tomo.SetCutFunction(plane)
                    alg_tomo.Update()
                    plane_sliced_tomo.shallow_copy(alg_tomo.GetOutput())
                    plane_widget.SetOrigin(origin)
                    plane_widget.SetNormal(normal)
                    plane_widget.UpdatePlacement()
            self.plt.add_key_event("Down", sliceDown)

            # Callback to place plane normal along X
            def sliceX():
                if self.tomo_slice_actor is not None:
                    alg_tomo = vtk.vtkCutter()
                    current_plane = vtk.vtkPlane()
                    alg_tomo.SetInputDataObject(self.pv_tomo_slice)
                    plane_sliced_tomo = self.plt.plane_sliced_meshes[-1]
                    plane_widget = self.plt.plane_widgets[-1]
                    plane_widget.GetPlane(current_plane)
                    origin = np.asarray(current_plane.GetOrigin())
                    normal = np.asarray([1, 0, 0])
                    # Create a new plane to update the position and perform the update
                    plane = generate_plane(normal, origin)
                    alg_tomo.SetCutFunction(plane)
                    alg_tomo.Update()
                    plane_sliced_tomo.shallow_copy(alg_tomo.GetOutput())
                    plane_widget.SetOrigin(origin)
                    plane_widget.SetNormal(normal)
                    plane_widget.UpdatePlacement()
            self.plt.add_key_event("x", sliceX)

            # Callback to place plane normal along Y
            def sliceY():
                if self.tomo_slice_actor is not None:
                    alg_tomo = vtk.vtkCutter()
                    current_plane = vtk.vtkPlane()
                    alg_tomo.SetInputDataObject(self.pv_tomo_slice)
                    plane_sliced_tomo = self.plt.plane_sliced_meshes[-1]
                    plane_widget = self.plt.plane_widgets[-1]
                    plane_widget.GetPlane(current_plane)
                    origin = np.asarray(current_plane.GetOrigin())
                    normal = np.asarray([0, 1, 0])
                    # Create a new plane to update the position and perform the update
                    plane = generate_plane(normal, origin)
                    alg_tomo.SetCutFunction(plane)
                    alg_tomo.Update()
                    plane_sliced_tomo.shallow_copy(alg_tomo.GetOutput())
                    plane_widget.SetOrigin(origin)
                    plane_widget.SetNormal(normal)
                    plane_widget.UpdatePlacement()
            self.plt.add_key_event("y", sliceY)

            # Callback to place plane normal along Z
            def sliceZ():
                if self.tomo_slice_actor is not None:
                    alg_tomo = vtk.vtkCutter()
                    current_plane = vtk.vtkPlane()
                    alg_tomo.SetInputDataObject(self.pv_tomo_slice)
                    plane_sliced_tomo = self.plt.plane_sliced_meshes[-1]
                    plane_widget = self.plt.plane_widgets[-1]
                    plane_widget.GetPlane(current_plane)
                    origin = np.asarray(current_plane.GetOrigin())
                    normal = np.asarray([0, 0, 1])
                    # Create a new plane to update the position and perform the update
                    plane = generate_plane(normal, origin)
                    alg_tomo.SetCutFunction(plane)
                    alg_tomo.Update()
                    plane_sliced_tomo.shallow_copy(alg_tomo.GetOutput())
                    plane_widget.SetOrigin(origin)
                    plane_widget.SetNormal(normal)
                    plane_widget.UpdatePlacement()
            self.plt.add_key_event("z", sliceZ)

            # Callback to reset plane origin
            def reserOrigin():
                if self.tomo_slice_actor is not None:
                    alg_tomo = vtk.vtkCutter()
                    current_plane = vtk.vtkPlane()
                    alg_tomo.SetInputDataObject(self.pv_tomo_slice)
                    plane_sliced_tomo = self.plt.plane_sliced_meshes[-1]
                    plane_widget = self.plt.plane_widgets[-1]
                    plane_widget.GetPlane(current_plane)
                    origin = np.asarray(self.pv_tomo_slice.center)
                    normal = np.asarray(np.asarray(current_plane.GetNormal()))
                    # Create a new plane to update the position and perform the update
                    plane = generate_plane(normal, origin)
                    alg_tomo.SetCutFunction(plane)
                    alg_tomo.Update()
                    plane_sliced_tomo.shallow_copy(alg_tomo.GetOutput())
                    plane_widget.SetOrigin(origin)
                    plane_widget.SetNormal(normal)
                    plane_widget.UpdatePlacement()
            self.plt.add_key_event("o", reserOrigin)

        if mask_mrc is not None:
            pos += 170. if pos != 0 else 45.
            self.plt.add_text('Vesicles', position=(pos, 65.), font_size=12)
            self.buttonGraph = self.plt.add_checkbox_button_widget(callback=self.plotMasks, position=(pos, 10.))

        if isinstance(self.points, np.ndarray):
            pos += 170. if pos != 0 else 45.
            self.plt.add_text('Coords', position=(pos, 65.), font_size=12)
            self.buttonPoints = self.plt.add_checkbox_button_widget(callback=self.plotPoints, position=(pos, 10.))

            if self.boxSize is not None:
                pos += 170.
                self.plt.add_text('Boxes', position=(pos, 65.), font_size=12)
                self.buttonBoxes = self.plt.add_checkbox_button_widget(callback=self.plotBoxes, position=(pos, 10.))

            # Picking Callbacks
            def removeSelection(selection):
                selected = selection['orig_extract_id']
                self.pv_points.remove_cells(selected, inplace=True)
                ids_removed = self.points_ids[selected]
                self.points_ids = np.delete(self.points_ids, selected)
                if self.normals is not None:
                    self.pv_normals = np.delete(self.pv_normals, selected, 0)
                    self.plt.remove_actor(self.normals_actor)
                    if self.buttonNormals.GetRepresentation().GetState():
                        self.normals_actor = self.plt.add_arrows(self.pv_points.cell_centers().points, self.pv_normals,
                                                                 mag=10, color='red', reset_camera=False)
                if self.buttonBoxes.GetRepresentation().GetState():
                    for point_id in ids_removed:
                        self.plt.remove_actor(self.box_actor[int(point_id - 1)])

            def enableRemoveSelection():
                self.plt.enable_cell_picking(mesh=self.pv_points,
                                             callback=removeSelection,
                                             font_size=12, opacity=0)

            # Picking Controls
            self.plt.main_menu.addAction('Point removal', enableRemoveSelection)

        if isinstance(self.normals, np.ndarray):
            pos += 170. if pos != 0 else 45.
            self.plt.add_text('Directions', position=(pos, 65.), font_size=12)
            self.buttonNormals = self.plt.add_checkbox_button_widget(callback=self.plotNormals, position=(pos, 10.))

    def getBinning(self, file):
        dim = ImageHandler().read(file + ':mrc').getDimensions()
        return int(np.floor(max(dim) / 400))

    def readMRC(self, file, binning=1, order=0, swapaxes=True):
        image = ImageHandler().read(file + ':mrc')
        data = np.squeeze(image.getData())
        data = zoom(data, 1 / (2 ** binning), order=order, prefilter=False)
        data = np.swapaxes(np.swapaxes(data, 0, 2), 1, 0) if swapaxes else data
        return data

    def gridFromMRC(self, data):
        '''Function to convert an MRC file into an Structure Grid in VTK'''

        # Create a Grid from the data
        x = np.arange(0, data.shape[0] + 1)
        y = np.arange(0, data.shape[1] + 1)
        z = np.arange(0, data.shape[2] + 1)
        x, y, z = np.meshgrid(x, y, z)

        # Copy the Grid to an Structure Grid VTK object (Pyvista)
        grid = pv.StructuredGrid(x, y, z)

        # Set the cell values. Previous reordering of the axis was needed to flatten properly the array
        grid.cell_arrays["values"] = data.flatten(order="K")

        return grid

    def surfaceFromMRC(self, data, label=1):
        '''Function to convert an MRC file into an Structure Surface in VTK'''

        # Get Only mesh corresponding to a given label (smooth the result to fill holes in the mask)
        data = data == label
        data = binary_erosion(binary_dilation(data, selem=ball(4)), selem=ball(1))

        # Triangulate coordinates using marching cubes algorithm
        grid = self.marchingCubes(data)

        # Fix the mesh
        mfix = pm._meshfix.PyTMesh(False)
        mfix.load_array(grid.points, grid.faces.reshape((-1, 4))[:, 1:])
        mfix.join_closest_components()
        mfix.fill_small_boundaries(refine=True)
        vert, faces = mfix.return_arrays()

        # Reconstruct fixed mesh
        triangles = np.empty((faces.shape[0], 4))
        triangles[:, -3:] = faces
        triangles[:, 0] = 3
        grid = pv.PolyData(vert, triangles.astype(int))

        # Final smoothing using pyvista
        grid = grid.smooth(n_iter=1000).clean()

        return grid

    def marchingCubes(self, volume, level=None, triangulation=True):
        vertices, faces = measure.marching_cubes(volume, spacing=(1, 1, 1), level=level, allow_degenerate=False)[:2]
        faces = np.column_stack((3 * np.ones((len(faces), 1), dtype=np.int), faces)).flatten()
        grid = pv.PolyData(vertices.astype(int), faces) if triangulation else pv.PolyData(vertices.astype(int))
        return grid

    def histogram(self, volume):
        hist, edges = np.histogram(volume, bins=100)
        hist = hist / np.sum(hist)
        bin_centers = np.mean(np.vstack([edges[0:-1], edges[1:]]), axis=0)
        return hist, bin_centers

    def contours(self, hist, bin_centers):
        logic_slicing = np.where((hist > np.std(hist)) * (hist < np.amax(hist)))
        sliced_hist = hist[logic_slicing]
        opacities = 1 - sliced_hist
        contour_values = bin_centers[logic_slicing]
        # print(1 - opacities)
        # print(np.mean(1-opacities))
        # print(contour_values)
        logic_slicing_2 = np.where(sliced_hist < np.mean(sliced_hist))
        contour_values = contour_values[logic_slicing_2]
        # print(contour_values)
        opacities = opacities[logic_slicing_2]
        return contour_values, opacities

    def isovolumes(self, volume, range=0.01, sigma=None, triangulation=True):
        volume = volume if sigma is None else gaussian_filter(volume, sigma=sigma)
        hist, bin_centers = self.histogram(volume)
        contour_values, opacities = self.contours(hist, bin_centers)
        opacities = (range / (np.amax(opacities) - np.amin(opacities))) * (opacities - np.amin(opacities))
        return [self.marchingCubes(volume, level, triangulation) for level in contour_values], opacities

    def downsamplingPC(self, coords, voxel_size):
        non_empty_voxel_keys, inverse, nb_pts_per_voxel = np.unique(((coords - np.min(coords, axis=0))
                                                                     // voxel_size).astype(int), axis=0,
                                                                    return_inverse=True,
                                                                    return_counts=True)
        idx_pts_vox_sorted = np.argsort(inverse)
        voxel_grid = {}
        grid_barycenter, grid_candidate_center = [], []
        last_seen = 0
        for idx, vox in enumerate(non_empty_voxel_keys):
            voxel_grid[tuple(vox)] = coords[idx_pts_vox_sorted[last_seen:last_seen + nb_pts_per_voxel[idx]]]
            grid_barycenter.append(np.mean(voxel_grid[tuple(vox)], axis=0))
            last_seen += nb_pts_per_voxel[idx]
        return np.asarray(grid_barycenter)

    def plotTomo(self, value):
        if value:
            # First check if tomogram is already loaded in memory, otherwise load it
            if isinstance(self.tomo, tuple):
                self.loading_screen.startAnimation()
                self.loadInMemory(source='tomo', sliceMode=False)
            else:
                self.showTomogram(load=False)
        else:
            for actor in self.tomo_actor:
                self.plt.remove_actor(actor)
            self.tomo_actor = []

    def toogleSlice(self, value):
        if value:
            # First check if tomogram is already loaded in memory, otherwise load it
            if isinstance(self.tomo, tuple):
                self.loading_screen.startAnimation()
                self.loadInMemory(source='tomo', sliceMode=True)
            else:
                self.showSlices(load=False)
        else:
            self.plt.remove_actor(self.tomo_slice_actor)
            self.plt.clear_plane_widgets()
            self.tomo_slice_actor = None

    def plotMasks(self, value):
        if value:
            # First check if mask is already loaded in memory, otherwise load it
            if isinstance(self.mask, str):
                self.loading_screen.startAnimation()
                self.loadInMemory(source='mask')
            else:
                self.showMask(load=False)
        else:
            for actor in self.mask_actors:
                self.plt.remove_actor(actor)
            self.graph_actor = []

    def plotPoints(self, value):
        if value:
            if self.first_reset:
                self.first_reset = False
                self.points_actor.append(self.plt.add_mesh(self.pv_points, show_scalar_bar=False, scalars="colors",
                                                           cmap="gist_rainbow_r", render_points_as_spheres=True, reset_camera=True))
            else:
                self.points_actor.append(self.plt.add_mesh(self.pv_points, show_scalar_bar=False, scalars="colors",
                                                           cmap="gist_rainbow_r", render_points_as_spheres=True, reset_camera=False))
        else:
            for actor in self.points_actor:
                self.plt.remove_actor(actor)
            self.points_actor = []

    def plotBoxes(self, value):
        if value:
            for point_id in self.points_ids:
                idp = int(point_id - 1)
                cube = pv.Cube(self.points[idp],
                               x_length=self.boxSize, y_length=self.boxSize, z_length=self.boxSize)
                self.box_actor[idp] = self.plt.add_mesh(cube, show_scalar_bar=False, style='wireframe',
                                                        color='blue')
        else:
            for actor in self.box_actor.values():
                self.plt.remove_actor(actor)
            self.box_actor = {}

    def plotNormals(self, value):
        if value:
            if self.first_reset:
                self.first_reset = False
                self.normals_actor = self.plt.add_arrows(self.pv_points.cell_centers().points, self.pv_normals,
                                                         mag=10, color='red', reset_camera=True)
            else:
                self.normals_actor = self.plt.add_arrows(self.pv_points.cell_centers().points, self.pv_normals,
                                                         mag=10, color='red', reset_camera=False)
        else:
            self.plt.remove_actor(self.normals_actor)
            self.normals_actor = None

    def initializePlot(self):
        # By default coordinates button is toggled
        if self.points is not None:
            self.buttonPoints.GetRepresentation().SetState(True)
            self.plotPoints(True)

        self.plt.app.exec_()

        # Save Points and Normals
        if self.save_basename is not None:
            np.savetxt(self.save_basename + '_indices.txt', self.points_ids, delimiter=' ')

    def loadInMemory(self, source, sliceMode=False):
        self.thread = QThread()
        self.worker = Worker(self)
        self.worker.moveToThread(self.thread)
        if source == 'tomo':
            self.thread.started.connect(self.worker.runTomoLoading)
            if sliceMode:
                self.thread.finished.connect(lambda: self.showSlices(load=True))
            else:
                self.thread.finished.connect(lambda: self.showTomogram(load=True))
        if source == 'mask':
            self.thread.started.connect(self.worker.runMaskLoading)
            self.thread.finished.connect(lambda: self.showMask(load=True))
        self.worker.finished.connect(self.thread.quit)
        # self.worker.finished.connect(self.worker.deleteLater)
        self.thread.finished.connect(self.thread.deleteLater)
        self.thread.start()

    def showTomogram(self, load=True):
        if load:
            self.tomo, self.pv_tomo, self.opacities, self.pv_tomo_slice = self.worker.output
            self.loading_screen.stopAnimation()
        cmap = matplotlib.cm.get_cmap('bone')  # Greys also looks nice
        cmap_ids = np.linspace(0, 1, len(self.pv_tomo))
        self.tomo_actor = [self.plt.add_mesh(actor, show_scalar_bar=False, opacity=3 * op, color=cmap(cid),
                                             render_points_as_spheres=True)
                           for actor, op, cid in zip(self.pv_tomo, self.opacities, cmap_ids)]
        self.plt.reset_camera()

    def showSlices(self, load=True):
        if load:
            self.tomo, self.pv_tomo, self.opacities, self.pv_tomo_slice = self.worker.output
            self.loading_screen.stopAnimation()
        self.tomo_slice_actor = self.plt.add_mesh_slice(self.pv_tomo_slice, normal='z', cmap="gray",
                                                        show_scalar_bar=False,
                                                        outline_translation=False, origin_translation=False)

    def showMask(self, load=True):
        if load:
            self.mask, self.pv_masks = self.worker.output
            self.loading_screen.stopAnimation()
        cmap = matplotlib.cm.get_cmap('Set3')
        cmap_ids = np.linspace(0, 1, len(self.pv_masks))
        self.mask_actors = [self.plt.add_mesh(mask, show_scalar_bar=False, color=cmap(cmap_id), smooth_shading=True)
                            for mask, cmap_id in zip(self.pv_masks, cmap_ids)]


class LoadingScreen(QWidget):
    def __init__(self):
        super().__init__()
        self.setFixedSize(200, 200)
        self.setWindowFlags(Qt.WindowStaysOnTopHint | Qt.CustomizeWindowHint)
        qtRectangle = self.frameGeometry()
        centerPoint = QDesktopWidget().availableGeometry().center()
        qtRectangle.moveCenter(centerPoint)
        self.move(qtRectangle.topLeft())

        radius = 20.0
        path = QtGui.QPainterPath()
        self.resize(440, 220)
        path.addRoundedRect(QRectF(self.rect()), radius, radius)
        mask = QtGui.QRegion(path.toFillPolygon().toPolygon())
        self.setMask(mask)

        self.label_animation = QLabel(self)
        plugin_path = os.path.dirname(tomoviz.__file__)
        self.movie = QMovie(os.path.join(plugin_path, "loading.gif"))
        self.label_animation.setMovie(self.movie)

    def startAnimation(self):
        self.movie.start()
        self.show()

    def stopAnimation(self):
        self.movie.stop()
        self.close()

class Worker(QObject):
    finished = pyqtSignal()

    def __init__(self, viewer):
        super().__init__()
        self.viewer = viewer

    def runTomoLoading(self):
        triangulation = self.viewer.tomo[1]
        sigma = self.viewer.tomo[2]
        tomo = self.viewer.readMRC(self.viewer.tomo[0], order=5, binning=self.viewer.binning)
        if isinstance(tomo, np.ndarray):
            pv_tomo_slice = self.viewer.gridFromMRC(tomo)
            pv_tomo, opacities = self.viewer.isovolumes(tomo, triangulation=triangulation, sigma=sigma)
        self.output = (tomo, pv_tomo, opacities, pv_tomo_slice)
        self.finished.emit()

    def runMaskLoading(self):
        mask = self.viewer.readMRC(self.viewer.mask, binning=self.viewer.binning)
        if isinstance(mask, np.ndarray):
            labels = np.unique(mask)[1:]
            pv_masks = [self.viewer.surfaceFromMRC(mask, label=label) for label in labels]
        self.output = (mask, pv_masks)
        self.finished.emit()



# **************************************************************************
# *
# * Authors:     David Herreros (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import numpy as np
from pyworkflow import utils as pwutils
from pyworkflow.gui.dialog import ToolbarListDialog
from pyworkflow.gui.tree import TreeProvider

from .viewer_triangulations import TriangulationPlot, guiThread
from .viewer_mrc import MrcPlot
from ..utils import delaunayTriangulation

from tomo.utils import extractVesicles, initDictVesicles, normalFromMatrix
import tomo.constants as const

class Tomo3DTreeProvider(TreeProvider):
    """ Populate Tree from SetOfTomograms. """

    def __init__(self, tomoList):
        TreeProvider.__init__(self)
        self.tomoList = tomoList

    def getColumns(self):
        return [('Tomogram', 300), ("# coords", 100)]

    def getObjectInfo(self, tomo):
        tomogramName = pwutils.removeBaseExt(tomo.getFileName())
        # tomogramName = pwutils.removeBaseExt(tomo.get())

        return {'key': tomogramName, 'parent': None,
                'text': tomogramName,
                'values':(tomo.count),
                'tags': ("done")}

    def getObjectPreview(self, obj):
        return (None, None)

    def getObjectActions(self, obj):
        return []

    def _getObjectList(self):
        """Retrieve the object list"""
        return self.tomoList

    def getObjects(self):
        objList = self._getObjectList()
        return objList

    def configureTags(self, tree):
        tree.tag_configure("done", foreground="black")


class Tomo3DDialog(ToolbarListDialog):
    """
    This class extend from ListDialog to allow calling
    a pyvista viewer subprocess from a list of Tomograms.
    """

    def __init__(self, parent, coordinates, extnormals, **kwargs):
        self.dictVesicles, _ = initDictVesicles(coordinates)
        self.coordinates = coordinates
        self.extnormals = extnormals
        self.provider = kwargs.get("provider", None)
        ToolbarListDialog.__init__(self, parent,
                                   "Tomogram List",
                                   allowsEmptySelection=False,
                                   itemDoubleClick=self.doubleClickOnTomogram,
                                   **kwargs)

    def doubleClickOnTomogram(self, e=None):
        self.tomo = e
        # tomoName = pwutils.removeBaseExt(self.tomo.getFileName())
        tomoName = pwutils.removeBaseExt(self.tomo.get())
        self.dictVesicles = extractVesicles(self.coordinates, self.dictVesicles, tomoName)
        self.createViewer()

    def createViewer(self):
        # tomoName = pwutils.removeBaseExt(self.tomo.getFileName())
        tomoName = pwutils.removeBaseExt(self.tomo.get())
        vesicles = self.dictVesicles[tomoName]['vesicles']
        shells = []
        for vesicle in vesicles:
            shells.append(delaunayTriangulation(vesicle))

        if self.extnormals is not None:
            normals = []
            extcoords = []
            normals.append([])
            for subtomo in self.extnormals:  # separate normals by tomoName/volId
                if pwutils.removeBaseExt(subtomo.getVolName()) == tomoName:
                    normal = normalFromMatrix(subtomo.getTransform().getMatrix())
                    coord = subtomo.getCoordinate3D()
                    coordSubtomo = [coord.getX(const.SCIPION),
                                    coord.getY(const.SCIPION),
                                    coord.getZ(const.SCIPION)]
                    normals[0].append(normal)
                    extcoords.append(coordSubtomo)

        else:
            normals = self.dictVesicles[tomoName]['normals']
            extcoords = None

        classArgs = {'meshes': shells, 'clouds': vesicles, 'extNormals_List': normals, 'extNormals_coords': extcoords}
        guiThread(TriangulationPlot, 'initializePlot', **classArgs)

class ViewerMRCDialog(ToolbarListDialog):
    """
    This class extend from ListDialog to allow calling
    a pyvista viewer subprocess from a list of Tomograms.
    """

    def __init__(self, parent, coords, **kwargs):
        self.coords = coords
        self.provider = kwargs.get("provider", None)
        ToolbarListDialog.__init__(self, parent,
                                   "Tomogram List",
                                   allowsEmptySelection=False,
                                   itemDoubleClick=self.doubleClickOnTomogram,
                                   **kwargs)

    def doubleClickOnTomogram(self, tomo=None):
        tomo_path = tomo.getFileName()
        coord_list = []
        direction_list = []
        for coord in self.coords.iterCoordinates(volume=tomo):
            direction = normalFromMatrix(coord.getMatrix())
            position = [coord.getX(const.BOTTOM_LEFT_CORNER),
                        coord.getY(const.BOTTOM_LEFT_CORNER),
                        coord.getZ(const.BOTTOM_LEFT_CORNER),
                        coord.getObjId()]
            position.append(coord.getGroupId()) if coord.getGroupId() is not None else position.append(0)
            coord_list.append(position)
            direction_list.append(direction)
        boxSize = self.coords.getBoxSize()
        np.savetxt('positions.txt', np.asarray(coord_list))
        np.savetxt('directions.txt', np.asarray(direction_list))
        viewer_args = {'tomo_mrc': tomo_path, 'points': 'positions.txt', 'normals': 'directions.txt',
                       'boxSize': boxSize}
        guiThread(MrcPlot, 'initializePlot', **viewer_args)



# **************************************************************************
# *
# * Authors:    Carlos Oscar Sorzano (coss@cnb.csic.es)
# *             Tomas Majtner (tmajtner@cnb.csic.es)  -- streaming version
# *             David Herreros Calero (dherreros@cnb.csic.es) -- Tomo version
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'coss@cnb.csic.es'
# *
# **************************************************************************
"""
Consensus picking protocol
"""

import numpy as np

from pyworkflow import BETA
import pyworkflow.protocol.params as params
from pyworkflow.utils import getFiles, removeBaseExt

from .protocol_particle_pick_consensus import (ProtTomoConsensusPicking,
                                               consensusWorker, getReadyTomos)

import tomo.constants as const


class ProtTomoPickingRemoveDuplicates(ProtTomoConsensusPicking):
    """
    This protocol removes coordinates that are closer than a given threshold.
    The remaining coordinate is the average of the previous ones.
    """

    _label = 'remove duplicates'
    _devStatus = BETA
    outputName = 'outputCoordinates'
    FN_PREFIX = 'purgedCoords_'

    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('inputCoordinates', params.PointerParam,
                      pointerClass='SetOfCoordinates3D',
                      label="Input coordinates", important=True,
                      help='Select the set of 3D coordinates to compare')
        form.addParam('consensusRadius', params.IntParam, default=10,
                      label="Radius",
                      help="All coordinates within this radius (in pixels) "
                           "are presumed to correspond to the same particle")

        # FIXME: It's not using more than one since
        #         self.stepsExecutionMode = STEPS_SERIAL
        # form.addParallelSection(threads=4, mpi=0)

#--------------------------- INSERT steps functions ----------------------------
    def insertNewCoorsSteps(self, tomos):
        deps = []
        for tomo in tomos:
            stepId = self._insertFunctionStep("removeDuplicatesStep",
                                              tomo.getObjId(),
                                              tomo.getFileName(),
                                              prerequisites=[])
            deps.append(stepId)
        return deps

    def _checkNewInput(self):
        # If continue from an stopped run, don't repeat what is done
        if not self.checkedTomos:
            for fn in getFiles(self._getExtraPath()):
                fn = removeBaseExt(fn)
                if fn.startswith(self.FN_PREFIX):
                    self.checkedTomos.update([self.getTomoId(fn)])
                    self.processedTomos.update([self.getTomoId(fn)])

        readyTomos, self.streamClosed = getReadyTomos(self.inputCoordinates.get())

        newTomosIds = readyTomos.difference(self.checkedTomos)

        if newTomosIds:
            self.checkedTomos.update(newTomosIds)

            inTomos = self.getMainInput().getPrecedents()
            newTomos = [inTomos[tomoId].clone() for tomoId in newTomosIds]

            fDeps = self.insertNewCoorsSteps(newTomos)
            outputStep = self._getFirstJoinStep()
            if outputStep is not None:
                outputStep.addPrerequisites(*fDeps)
            self.updateSteps()

    def getMainInput(self):
        return self.inputCoordinates.get()

    def defineRelations(self, outputSet):
        self._defineTransformRelation(self.getMainInput(), outputSet)

    def removeDuplicatesStep(self, tomoId, tomoName):
        print("Removing duplicates for tomogram %d: '%s'"
              % (tomoId, tomoName))

        coordArray = np.asarray([x.getPosition(const.SCIPION) for x in
                                 self.getMainInput().iterCoordinates(tomoId)],
                                 dtype=int)
        vIds = np.asarray([x.getGroupId() for x in
                           self.getMainInput().iterCoordinates(tomoId)],
                           dtype=int)
        trMat = np.asarray([x.getMatrix() for x in
                            self.getMainInput().iterCoordinates(tomoId)])

        consensusWorker([coordArray], [vIds], [trMat], 1, self.consensusRadius.get(),
                        self._getTmpPath('%s%s.txt' % (self.FN_PREFIX, tomoId)))

        self.processedTomos.update([tomoId])

    def _validate(self):
        errors = []
        return errors

    def _summary(self):
        return ["Radius = %d" % self.consensusRadius]



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
from tomoviz import __version__

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='scipion-em-tomoviz',  # Required
    version=__version__,  # Required
    description='Tools for 3D visualization and manipulation of tomography data',  # Required
    long_description=long_description,  # Optional
    url='https://github.com/scipion-em/scipion-em-tomoviz',  # Optional
    author='you',  # Optional
    author_email='you@yourinstitution.email',  # Optional
    keywords='scipion tomography imageprocessing scipion-3.0',  # Optional
    packages=find_packages(),
    install_requires=[requirements],
    entry_points={'pyworkflow.plugin': 'tomoviz = tomoviz'},
    package_data={  # Optional
       'tomoviz': ['icon.png', 'icon_square.png', 'loading.gif', 'protocols.conf'],
    }
)



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import sys
import pyvista as pv
import numpy as np
from scipy.spatial import cKDTree


def rotation_matrix_from_vectors(vec1, vec2):
    # a, b = (vec1 / np.linalg.norm(vec1)).reshape(3), (vec2 / np.linalg.norm(vec2)).reshape(3)
    a, b = vec1, vec2
    v = np.cross(a, b)
    c = np.dot(a, b)
    s = np.linalg.norm(v)
    if s != 0:
        tr = np.zeros([4, 4])
        kmat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
        rotation_matrix = np.eye(3) + kmat + kmat.dot(kmat) * ((1 - c) / (s ** 2))
        tr[0:3, 0:3] = rotation_matrix
        tr[-1, -1] = 1
        return tr
    else:
        return np.eye(4)

def inverse_Transformation(transformation):
    translation = np.eye(4)
    rotation = np.eye(4)
    translation[0:-1, -1] = -transformation[0:-1, -1]
    rotation[:3, :3] = transformation[:3, :3]
    return np.linalg.inv(rotation).dot(translation)

def delaunayTriangulation(cloud, adjustCloud=True):
    cloud_pv = pv.PolyData(cloud)
    cloud_pv.GlobalWarningDisplayOff()
    mesh = cloud_pv.delaunay_3d()
    shell = mesh.extract_geometry().triangulate()

    # If points lie outside the shell, they are readjusted to compute their normals
    if adjustCloud:
        shell.subdivide(4, inplace=True)
        areZero = np.where((shell.point_normals == (0, 0, 0)).all(axis=1))
        points = np.asarray(shell.points)
        # points[areZero] = np.array((0, 0, 0))
        points = np.delete(points, areZero, axis=0)
        tree = cKDTree(points, leafsize=100)
        _, idc = tree.query(cloud)
        newCoords = points[idc]
        # newCoords = np.asarray([points[np.argmin(np.linalg.norm(points - point, axis=1))]
        #                         for point in cloud])
        shell = delaunayTriangulation(newCoords, adjustCloud=False)
    return shell

def computeNormals(triangulation, associateCoords=False):
    triangulation.compute_normals(inplace=True)
    normals = triangulation.point_normals
    points = triangulation.points

    # Check if coordinates are repeated (due to neighbour search) and copy the normal
    # so it is not (0,0,0)
    _, unique_indices, unique_inverse = np.unique(points, return_index=True, return_inverse=True, axis=0)
    unique_normals = normals[unique_indices]
    normals = unique_normals[unique_inverse]

    # NOT USED
    # for i in range(len(points)):  # generate pairs
    #     for j in range(i + 1, len(points)):
    #         if np.array_equal(points[i], points[j]):  # compare rows
    #             normals[j] = normals[i]
    #         else:
    #             pass

    # Sometimes, points may be redundant to the mesh
    # Assign the closest normal to them
    # Poosible cKDTree?
    areZero = np.where((normals == (0, 0, 0)).all(axis=1))
    if areZero[0].size != 0:
        redundant = points[areZero]
        points = np.delete(points, areZero, axis=0)
        ngNormals = np.asarray([np.argmin(np.linalg.norm(points - point, axis=1))
                                for point in redundant])
        normals[areZero] = normals[ngNormals]

    if associateCoords:
        return [(np.asarray(point), np.asarray(normal)) for point, normal in zip(points, normals)]
    else:
        return np.asarray(normals)



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import os.path

import numpy as np
import pyworkflow.viewer as pwviewer
from pyworkflow.object import String
from pyworkflow.gui.dialog import askYesNo
from pyworkflow.utils.properties import Message
import pyworkflow.utils as pwutils

from pwem.protocols import EMProtocol
import pwem.viewers.views as vi
from .views_tkinter_tree import Tomo3DTreeProvider
from .views_tkinter_tree import Tomo3DDialog, ViewerMRCDialog

import tomo.objects
from ..protocols import XmippProtFilterbyNormal


class TomoVizDataViewer(pwviewer.Viewer):
    """ Wrapper to visualize different type of objects
    using pyvista
    """
    _environments = [pwviewer.DESKTOP_TKINTER]
    _targets = [
        tomo.objects.SetOfCoordinates3D,
        XmippProtFilterbyNormal
    ]

    def __init__(self, **kwargs):
        pwviewer.Viewer.__init__(self, **kwargs)
        self._views = []

    def _getObjView(self, obj, fn, viewParams={}):
        return vi.ObjectView(
            self._project, obj.strId(), fn, viewParams=viewParams)

    def _visualize(self, obj, **kwargs):
        views = []
        cls = type(obj)

        if issubclass(cls, tomo.objects.SetOfCoordinates3D):
            outputCoords = obj
        elif issubclass(cls, EMProtocol):
            outputCoords = obj.inputMeshes.get()

        tomos = outputCoords.getPrecedents()

        volIds = outputCoords.aggregate(["MAX", "COUNT"], "_volId", ["_volId"])
        volIds = [(d['_volId'], d["COUNT"]) for d in volIds]

        tomoList = []
        for objId in volIds:
            tomogram = tomos[objId[0]].clone()
            tomogram.count = objId[1]
            tomoList.append(tomogram)
                        # tomoList = [String(tomos[objId].getFileName()) for objId in volIds]
        tomoProvider = Tomo3DTreeProvider(tomoList)
        ViewerMRCDialog(self._tkRoot, outputCoords, provider=tomoProvider)

        import tkinter as tk
        frame = tk.Frame()
        if askYesNo(Message.TITLE_SAVE_OUTPUT, Message.LABEL_SAVE_OUTPUT, frame):
            protocol = self.protocol
            suffix = protocol._getOutputSuffix(tomo.objects.SetOfCoordinates3D)
            updated_set = protocol._createSetOfCoordinates3D(tomos, suffix)
            updated_set.setName("Selected Coordinates")
            updated_set.setPrecedents(tomos)
            updated_set.setSamplingRate(tomos.getSamplingRate())
            updated_set.setBoxSize(outputCoords.getBoxSize())
            for item in tomoList:
                basename = pwutils.removeBaseExt(item.getFileName())
                indices_file = basename + '_indices.txt'
                if os.path.isfile(indices_file):
                    indices = np.loadtxt(indices_file, delimiter=' ')
                    for index in indices:
                        updated_set.append(outputCoords[index].clone())
                    pwutils.cleanPath(indices_file)
            name = protocol.OUTPUT_PREFIX + suffix
            args = {}
            args[name] = updated_set
            protocol._defineOutputs(**args)
            protocol._defineSourceRelation(tomos, updated_set)
            protocol._updateOutputSet(name, updated_set, state=updated_set.STREAM_CLOSED)
        else:
            for item in tomoList:
                basename = pwutils.removeBaseExt(item.getFileName())
                indices_file = basename + '_indices.txt'
                if os.path.isfile(indices_file):
                    pwutils.cleanPath(indices_file)

        return views
